import numpy
from ctypes import *
from typing import Dict, List, Union
from tract.bindings import check, lib

TRACT_DATUM_TYPE_BOOL = 0x01
TRACT_DATUM_TYPE_U8 = 0x11
TRACT_DATUM_TYPE_U16 = 0x12
TRACT_DATUM_TYPE_U32 = 0x14
TRACT_DATUM_TYPE_U64 = 0x18
TRACT_DATUM_TYPE_I8 = 0x21
TRACT_DATUM_TYPE_I16 = 0x22
TRACT_DATUM_TYPE_I32 = 0x24
TRACT_DATUM_TYPE_I64 = 0x28
TRACT_DATUM_TYPE_F16 = 0x32
TRACT_DATUM_TYPE_F32 = 0x34
TRACT_DATUM_TYPE_F64 = 0x38
TRACT_DATUM_TYPE_COMPLEX_I16 = 0x42
TRACT_DATUM_TYPE_COMPLEX_I32 = 0x44
TRACT_DATUM_TYPE_COMPLEX_I64 = 0x48
TRACT_DATUM_TYPE_COMPLEX_F16 = 0x52
TRACT_DATUM_TYPE_COMPLEX_F32 = 0x54
TRACT_DATUM_TYPE_COMPLEX_F64 = 0x58

def dt_numpy_to_tract(dt):
    if dt.kind == 'b':
        return TRACT_DATUM_TYPE_BOOL
    if dt.kind == 'u':
        return 0x10 + dt.itemsize
    if dt.kind == 'i':
        return 0x20 + dt.itemsize
    if dt.kind == 'f':
        return 0x30 + dt.itemsize
    if dt.kind == 'c':
        return 0x50 + dt.itemsize / 2
    raise TractError("Unsupported Numpy dtype: " + dt)


class Value:
    """
    Represents a tensor suitable for manipulation by tract.

    On the Python side, the main way to access tensor data is to
    convert the value to a numpy array.
    """
    def __init__(self, ptr):
        self.ptr = ptr

    def __del__(self):
        if self.ptr:
            check(lib.tract_value_destroy(byref(self.ptr)))

    def _valid(self):
        if self.ptr == None:
            raise TractError("invalid value (maybe already consumed ?)")

    def from_numpy(array: numpy.ndarray) -> "Value":
        array = numpy.ascontiguousarray(array)

        data = array.__array_interface__['data'][0]
        data = c_void_p(data)
        ptr = c_void_p()
        shape_t = c_size_t * array.ndim
        shape = shape_t()
        for ix in range(0, array.ndim):
            shape[ix] = array.shape[ix]
        dt = dt_numpy_to_tract(array.dtype)
        check(lib.tract_value_from_bytes(dt, c_size_t(array.ndim), shape, data, byref(ptr)))
        return Value(ptr)

    def to_numpy(self) -> numpy.array:
        """Builds a numpy array equivalent to the data in this value."""
        self._valid()
        rank = c_size_t();
        shape = POINTER(c_size_t)()
        dt = c_float
        data = POINTER(dt)()
        check(lib.tract_value_as_bytes(self.ptr, None, byref(rank), byref(shape), byref(data)))
        rank = rank.value
        shape = [ int(shape[ix]) for ix in range(0, rank) ]
        array = numpy.ctypeslib.as_array(data, shape).copy()
        return array

    def into_numpy(self) -> numpy.array:
        """Same as to_numpy(), but drop the value content once the numpy array is built."""
        result = self.to_numpy()
        check(lib.tract_value_destroy(byref(self.ptr)))
        return result




from ctypes import *
from pathlib import Path

if len(list(Path(__file__).parent.glob("*.so"))) > 0:
    dylib_path = list(Path(__file__).parent.glob("*.so"))[0]
elif len(list(Path(__file__).parent.glob("*.pyd"))) > 0:
    dylib_path = list(Path(__file__).parent.glob("*.pyd"))[0]
else:
    raise TractError("Can not find dynamic library")

lib = cdll.LoadLibrary(str(dylib_path))

lib.tract_version.restype = c_char_p
lib.tract_get_last_error.restype = c_char_p
lib.tract_free_cstring.restype = None

class TractError(Exception):
    pass

def check(err):
    if err != 0:
        raise TractError(str(lib.tract_get_last_error(), "utf-8"))




import numpy
from ctypes import *
from typing import Dict, List, Union
from .bindings import check, lib
from .fact import Fact
from .value import Value

class State:
    """
    The state of a stateful model.
    """
    def __init__(self, ptr):
        self.ptr = ptr

    def __del__(self):
        check(lib.tract_state_destroy(byref(self.ptr)))

    def _valid(self):
        if self.ptr == None:
            raise TractError("invalid state (maybe already destroyed ?)")

    def input_count(self) -> int:
        """Return the number of inputs of the underlying model"""
        self._valid()
        i = c_size_t()
        check(lib.tract_state_input_count(self.ptr, byref(i)))
        return i.value

    def output_count(self) -> int:
        """Return the number of outputs of the underlying model"""
        self._valid()
        i = c_size_t()
        check(lib.tract_state_output_count(self.ptr, byref(i)))
        return i.value

    def run(self, inputs: List[Union[Value, numpy.ndarray]]) -> List[Value]:
        """
        Runs the model over the provided input list, and returns the model outputs.
        """
        self._valid()
        input_values = []
        for v in inputs:
            if isinstance(v, Value):
                input_values.append(v)
            elif isinstance(v, numpy.ndarray):
                input_values.append(Value.from_numpy(v))
            else:
                raise TractError(f"Inputs must be of type tract.Value or numpy.Array, got {v}")
        input_ptrs = (c_void_p * self.input_count())()
        output_ptrs = (c_void_p * self.output_count())()
        for ix, v in enumerate(input_values):
            input_ptrs[ix] = v.ptr
        check(lib.tract_state_run(self.ptr, input_ptrs, output_ptrs))
        result = []
        for v in output_ptrs:
            result.append(Value(c_void_p(v)))
        return result

    def freeze(self) -> "FrozenState":
        self._valid()
        frozen = c_void_p()
        check(lib.tract_state_freeze(self.ptr, byref(frozen)))
        return FrozenState(frozen)

class FrozenState:
    """
    The state of a stateful model.
    """
    def __init__(self, ptr):
        self.ptr = ptr

    def __del__(self):
        check(lib.tract_frozen_state_destroy(byref(self.ptr)))

    def _valid(self):
        if self.ptr == None:
            raise TractError("invalid frozen state (maybe already destroyed ?)")

    def unfreeze(self) -> State:
        self._valid()
        state = c_void_p()
        check(lib.tract_frozen_state_unfreeze(self.ptr, byref(state)))
        return State(state)



import numpy
from ctypes import *
from typing import Dict, List, Union
from .bindings import check, lib
from .fact import Fact
from .value import Value
from .runnable import Runnable

class Model:
    """
    # Main model object

    ## Central focus point of the model transformation pipeline
 
    The Model is the central point of tract model loading and "model cooking". ONNX and NNEF 
    serialized models are converted to Model (more or less directly) before we can do anything
    of value with them. Model can be dumped to NNEF (or tract-opl which is NNEF plus tract
    proprietary extensions).
    
    A Model can be `optimize()`, substituing the "high level" operators in tract-core operator set by
    the best implementation available for the current system. From there it can be transformed into a 
    Runnable object that we will use to run.
 
    ## Model cooking
 
    But some model transformations can be performed on the Model class:
 
    - declutter (getting rid of training artefacts)
    - "pulsification" (transforming a batch-oriented model into a streaming model)
    - symbol substitution (make N or Batch a fixed number, unlocking potential optimisation later on)
    - static cost evalation and dynamic profiling
    - ...
 
    In some situation, these operation are done "on-the-fly" when a ONNX or NNEF model is loaded,
    at start-up time. In other situation, when start-up time becomes an issue, it may be beneficial
    to "pre-cook" the model: apply the transformations one time, serialize the model as NNEF (with
    tract-opl extension if needed). At start-up this model can be significantly less expensive to
    "cook" for inference.
 
    ## Model and TypedModel

    This class is actually a wrapper around the "TypedModel" in Rust codebase. The "Typed"
    bit means than all shapes and element types in all input, output and temporary values must
    known. There is support in tract for symbols in dimensions, with some limited computation
    capabilities on symbolic expression. For instance, it is relatively frequent to work with
    a Model where all tensors shapes start with the `N` or `Batch`.
    """

    def __init__(self, ptr):
        self.ptr = ptr

    def __del__(self):
        if self.ptr:
            check(lib.tract_model_destroy(byref(self.ptr)))

    def _valid(self):
        if self.ptr == None:
            raise TractError("invalid model (maybe already consumed ?)")

    def input_count(self) -> int:
        """Return the number of inputs of the model"""
        self._valid()
        i = c_size_t()
        check(lib.tract_model_input_count(self.ptr, byref(i)))
        return i.value

    def output_count(self) -> int:
        """Return the number of outputs of the model"""
        self._valid()
        i = c_size_t()
        check(lib.tract_model_output_count(self.ptr, byref(i)))
        return i.value

    def input_name(self, input_id: int) -> str:
        """Return the name of the input_id-th input"""
        self._valid()
        cstring = c_char_p()
        check(lib.tract_model_input_name(self.ptr, input_id, byref(cstring)))
        result = str(cstring.value, "utf-8")
        lib.tract_free_cstring(cstring)
        return result

    def input_fact(self, input_id: int) -> Fact:
        """Return the fact of the input_id-th input"""
        self._valid()
        fact = c_void_p()
        check(lib.tract_model_input_fact(self.ptr, input_id, byref(fact)))
        return Fact(fact)

    def set_output_names(self, names: List[str]):
        """Change the output nodes of the model"""
        self._valid()
        nb = len(names)
        names_str = []
        names_ptr = (c_char_p * nb)()
        for ix, n in enumerate(names):
            names_str.append(str(n).encode("utf-8"))
            names_ptr[ix] = names_str[ix]
        check(lib.tract_model_set_output_names(self.ptr, nb, names_ptr))

    def output_name(self, output_id: int) -> str:
        """Return the name of the output_id-th output"""
        self._valid()
        cstring = c_char_p()
        check(lib.tract_model_output_name(self.ptr, output_id, byref(cstring)))
        result = str(cstring.value, "utf-8")
        lib.tract_free_cstring(cstring)
        return result

    def output_fact(self, input_id: int) -> Fact:
        """Return the fact of the output_id-th output"""
        self._valid()
        fact = c_void_p()
        check(lib.tract_model_output_fact(self.ptr, input_id, byref(fact)))
        return Fact(fact)

    def concretize_symbols(self, values: Dict[str, int]) -> None:
        """Substitute symbols by a value

        Replace all occurencies of the symbols in the dictionary, in all the Model facts shapes.

        While this is not strictly necesary, the optimizing steps may make better choices if the model
        is informed of some specific symbol values.
        """
        self._valid()
        nb = len(values)
        names_str = []
        names = (c_char_p * nb)()
        values_list = (c_int64 * nb)()
        for ix, (k, v) in enumerate(values.items()):
            names_str.append(str(k).encode("utf-8"))
            names[ix] = names_str[ix]
            values_list[ix] = v
        check(lib.tract_model_concretize_symbols(self.ptr, c_size_t(nb), names, values_list))

    def pulse(self, symbol: str, pulse: Union[str, int]) -> None:
        """Pulsify a model.

        `pulse` is typically a one-length dictionary mapping the time dimension symbol to a pulse len.
        """
        self._valid()
        check(lib.tract_model_pulse_simple(byref(self.ptr), symbol.encode("utf-8"), str(pulse).encode("utf-8")))

    def half(self) -> None:
        """Convert the model to hald precision
        """
        self._valid()
        check(lib.tract_model_half(self.ptr))

    def declutter(self) -> None:
        """Declutter a model.

        Perform the first "half" of optimisation phases, consisting or removing training artefacts and converge
        on tract-core canonical form.
        """
        self._valid()
        check(lib.tract_model_declutter(self.ptr))

    def optimize(self) -> None:
        """Optimize a model.

        Perform the second "half" of optimisation phases, consisting of translating the tract-core canonical
        form to the best runtime for the current architecture.
        """
        self._valid()
        check(lib.tract_model_optimize(self.ptr))

    def into_decluttered(self) -> "Model":
        """Convenience method performing `declutter()` and returning the model"""
        self.declutter();
        return self

    def into_optimized(self) -> "Model":
        """Convenience method performing `optimize()` and returning the model"""
        self.optimize()
        return self

    def into_runnable(self) -> Runnable:
        """Transform the model into a ready to be used Runnable model"""
        self._valid()
        runnable = c_void_p()
        check(lib.tract_model_into_runnable(byref(self.ptr), byref(runnable)))
        return Runnable(runnable)

    def property_keys(self) -> List[str]:
        """Query the list of properties names of the model."""
        self._valid()
        count = c_size_t()
        check(lib.tract_model_property_count(self.ptr, byref(count)))
        count = count.value
        cstrings = (POINTER(c_char) * count)()
        check(lib.tract_model_property_names(self.ptr, cstrings))
        names = []
        for i in range(0, count):
            names.append(str(cast(cstrings[i], c_char_p).value, "utf-8"))
            lib.tract_free_cstring(cstrings[i])
        return names

    def property(self, name: str) -> Value:
        """Query a property by name"""
        self._valid()
        value = c_void_p()
        check(lib.tract_model_property(self.ptr, str(name).encode("utf-8"), byref(value)))
        return Value(value)

    def profile_json(self, inputs: Union[None, List[Union[Value, numpy.ndarray]]]) -> str:
        """Profile the model. Also compute the static costs of operators.

        Returns is a json buffer.
        """
        self._valid()
        cstring = c_char_p()
        input_values = []
        input_ptrs = None
        if inputs != None:
            for v in inputs:
                if isinstance(v, Value):
                    input_values.append(v)
                elif isinstance(v, numpy.ndarray):
                    input_values.append(Value.from_numpy(v))
                else:
                    raise TractError(f"Inputs must be of type tract.Value or numpy.Array, got {v}")
            input_ptrs = (c_void_p * len(inputs))()
            for ix, v in enumerate(input_values):
                input_ptrs[ix] = v.ptr
        check(lib.tract_model_profile_json(self.ptr, input_ptrs, byref(cstring)))
        result = str(cstring.value, "utf-8")
        lib.tract_free_cstring(cstring)
        return result




"""
`tract` Python bindings library

`tract` is a neural network inference engine.
Its main purpose is to *run* a neural network on production premises after it has been trained.
It is a native library written in Rust, and specific attention has been given to its performance
in sound streaming applications in embedded context (typically ARM Cortex-A CPUs), but it is meant
to be a generic engine, performing more than decently with various loads on most architectures. You
can use it to run image categorization on a PC.

```python
import tract

# load MobileNet version 2, an image categorization model
model = (
    tract.onnx()
    .model_for_path("./mobilenetv2-7.onnx")
    .into_optimized()
    .into_runnable()
)

# load, as a numpy array, a picture of Grace Hopper, wearing her military uniform
grace_hopper_1x3x224x244 = numpy.load("grace_hopper_1x3x224x244.npy")

# run the image through Mobilenet in tract
result = model.run([grace_hopper_1x3x224x244])

# output is an array of confidence for each class of the ImageNet challenge
confidences = result[0].to_numpy()

# class 652 is "military uniform"
assert numpy.argmax(confidences) == 652
```

`tract` can also be used as a "model cooking" toolbox: once a model has been trained, it is
sometimes useful to perform some transformations, simplifications and optimizations before shipping
it. These bindings offer access to some of `tract` cooking facilities.
"""

import numpy
from ctypes import *
from pathlib import Path
from typing import Dict, List, Union

from .bindings import check, lib, TractError
from .value import Value
from .fact import Fact, InferenceFact
from .model import Model
from .inference_model import InferenceModel
from .runnable import Runnable
from .nnef import Nnef
from .onnx import Onnx

def version() -> str:
    """Return the version string of `tract` native library"""
    return str(lib.tract_version(), "utf-8")

def nnef() -> Nnef:
    """Return a newly-created NNEF context for loading and saving models"""
    return Nnef()

def onnx() -> Onnx:
    """Return a newly-created ONNX context for loading models"""
    return Onnx()

